/**
 *	@file bt.c
 *
 *	@project btChat
 *
 *	@author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 * (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "util.h"
#include "discover_protocol.h"
#include "node_protocol.h"
#include "event.h"

/**
 *	@brief wait for incoming bluetooth connections
 *
 *
 *	@returns -1 on error, 1 on success and 2 with penning or generated events
 */
int bt_server(struct btchatd_t *globals, struct pollfd *fd)
{
	struct sockaddr_rc remote_sa = {0};
	int remote_len = 0;
	int remote_socket = -1;
	int i;
	int tmp = 0;
	int ins_pos = -1;
	int res;
	

	#ifdef DEBUG
	assert(globals != NULL);
	#endif

	
	/* init btserver */
	if (globals->btserver.state == 0) {
		if ((globals->btserver.server = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
			#ifdef DEBUG
			perror("bt_server(): socket");
			#endif
			return(-1);
		}
		globals->btserver.server_sa.rc_family = AF_BLUETOOTH;
		globals->btserver.server_sa.rc_channel = globals->btserver.channel;
		globals->btserver.server_sa.rc_bdaddr = globals->btserver.local_bdaddr;
		
		if (bind(globals->btserver.server, (struct sockaddr*) &globals->btserver.server_sa, sizeof(globals->btserver.server_sa)) < 0) {
			#ifdef DEBUG
			perror("bt_server(): bind");
			#endif
			return(-1);
		}
	
		if (listen(globals->btserver.server, 1) < 0) {
			#ifdef DEBUG
			perror("bt_server(): listen");
			#endif
			return(-1);
		}
			
		globals->btserver.state = 1;
		return(1);
	}
	/* ready to accept new connections */
	else if (globals->btserver.state == 1) {
	
		/* check polled events */
		if ((fd->revents & POLLIN) != POLLIN && (fd->revents & POLLPRI) != POLLPRI) return(1);
	
		if (fd->fd == globals->btserver.server) {
			remote_socket = accept(globals->btserver.server, (struct sockaddr*) &remote_sa, &remote_len);
			if (remote_socket != -1) {
				/* only allow connection if we are below the limit */
				if (globals->btserver.clients_max > globals->btserver.clients_num) {
					/* make sure that we don't have a connection to this client all ready */
					for (i = 0; i < 7; i++) {
						if (globals->btserver.clients[i].s != -1) {
							if (memcmp(&globals->btserver.clients[i].sa.rc_bdaddr, &remote_sa.rc_bdaddr, sizeof(bdaddr_t) == 0)) {
								tmp = 1;
								break;
							}
						}
						else {
							/* remember insertion position for new client */
							if (ins_pos == -1) ins_pos = i;
						}
					} /* for () */
					
					if (tmp == 0) {
						/* setup new client */
						res = bt_client_direct_init(globals, ins_pos, remote_socket, &remote_sa, remote_len);
						if (res == 1) {
							globals->btserver.clients_num++;
							bt_notify_net(globals, NULL, ins_pos);
							return(2);
						}
						else {
							globals->btserver.clients[ins_pos].s = -1;
							close(remote_socket);
							return(1);
						}
					}
					else {
						#ifdef DEBUG_INFO
						printf("bt_server(): all ready connected\n");
						#endif
						close(remote_socket);
					}
				}
				else {
					#ifdef DEBUG_INFO
					printf("bt_server(): too many connections\n");
					#endif
					close(remote_socket);
				}
			}
			else {
				#ifdef DEBUG
				perror("bt_server(): accept\n");
				#endif
				return(-1);
			}
		}
		else {
			#ifdef DEBUG
			printf("bt_server(): fd.fd has strange value!\n");
			#endif
			
			/* set state to self init */
			globals->btserver.state = 0;
			/* indicate penning events */
			return(2);
		}
	}
	/* stop accepting new connections */
	else if (globals->btserver.state == 2) {
		if (globals->btserver.server != -1) {
			close(globals->btserver.server);
			globals->btserver.server = -1;
		}
	}
	
	return(1);
}

/**
 *	@brief handle client connection
 *
 *	@param globals	pointer to globals
 *	@param fd	pointer to poll event
 *	@param index	index of client connection in btserver.clients[X] array
 *
 *	@returns -1 on error 1 on success and 2 on pending or generated events
 */
int bt_client(struct btchatd_t *globals, struct pollfd *fd, int index)
{
	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	

	if (globals->btserver.clients[index].state == 0) {
		#ifdef DEBUG_INFO
		/* printf("bt_client[%d](): state=0, should not happen!\n", index); */
		#endif
	}
	else if (globals->btserver.clients[index].state == 1) {
		if (fd->fd == globals->btserver.clients[index].s) {
			if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
				/* read data */
			}
		}
		else {
			#ifdef DEBUG
			printf("bt_client[%d](): fd->fd has strange value\n", index);
			#endif
			
			/*
			 * close connection
			 */
			
			return(-1);
		}
	}
	else {
		#ifdef DEBUG_INFO
		printf("bt_client[%d](): unknown state=%d\n", index, globals->btserver.clients[index].state);
		#endif
	}

	return(1);
}

int bt_client_connect(struct btchatd_t *globals)
{
	return(1);
}

/**
 *	@brief talk to bluetooth device discovery server
 *
 *	@param globals	pointer to globals
 *	@param fd	pointer to poll events
 *
 *	@returns -1 on error 1 on success and 2 on pending or generated events
 */
int bt_discover_client(struct btchatd_t *globals, struct pollfd *fd)
{
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_device_t *dev;
	struct btcdiscover_protocol_discover_t *dis;
	int read_bytes = 0;
	struct btchat_event_t *event = NULL;
	unsigned char out_buffer[1024];
	struct btchat_node_t node;
	struct btchat_node_t *new_node;
	struct tree_node_t *result_node;
	int i;
	int res;
	int num_devices;
	#ifdef DEBUG
	char addr_str[20];
	#endif
	int ret_val = 1;
		
	
	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	
	
	if (fd == NULL) {
		#ifdef DEBUG
		printf("bt_discover_client(): fd == NULL\n");
		#endif
		return(-1);
	}
		
	if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
	
		read_bytes = cread(globals->btdiscover.s, globals->btdiscover.buffer + globals->btdiscover.buffer_pos, globals->btdiscover.buffer_size - globals->btdiscover.buffer_pos);		
		if (read_bytes == 0) {
			#ifdef DEBUG
			printf("bt_discover_client(): read_bytes=0, looks like the bt_discover_server() closed the socket, terminate\n");
			#endif
			return(1);
		}
		globals->btdiscover.buffer_pos += read_bytes;

	check_buffer_again:			
		if (globals->btdiscover.buffer_pos >= sizeof(struct btcdiscover_protocol_header_t)) {
			he = (struct btcdiscover_protocol_header_t*) globals->btdiscover.buffer;
				
			#ifdef DEBUG_INFO
			printf("bt_discover_client(): he->length=%d, buffer_pos=%d\n", he->length, globals->btdiscover.buffer_pos);
			#endif
				
			if (globals->btdiscover.buffer_pos >= he->length) {
			
				/* device info packet */
				if (he->pkg_type == btcdiscover_protocol_device) {
					num_devices = (he->length - sizeof(struct btcdiscover_protocol_header_t)) / sizeof(struct btcdiscover_protocol_device_t);
						
					#ifdef DEBUG_INFO
					printf("bt_discover_client(): %d devices\n", num_devices);
					#endif
						
					for (i = 0; i < num_devices; i++) {
						dev = (struct btcdiscover_protocol_device_t*) (globals->btdiscover.buffer + sizeof(struct btcdiscover_protocol_header_t) + (i * sizeof(struct btcdiscover_protocol_device_t)));
							
						memset(&node, 0, sizeof(struct btchat_node_t));
						memcpy(node.addr.b, dev->addr, 6);

						result_node = NULL;							
						res = tree_access_insert(&globals->nodes, &node, &result_node);
						
						switch (res) {
						case 1:	/* known node */
							#ifdef DEBUG_INFO
							ba2str(&node.addr, addr_str);
							printf("bt_discover_client(): known node dev=%d bdaddr=%s\n", i, addr_str);
							#endif
							new_node = (struct btchat_node_t*) result_node->data;
							
							if (new_node->state != dev->state) {
								#ifdef DEBUG_INFO
								printf("bt_discover_client(): known node state old=%d new=%d\n", new_node->state, dev->state);
								#endif
								
								new_node->state = dev->state;
								memcpy(new_node->name, dev->name, 250);
								new_node->device_class = dev->device_class;
								new_node->btchat_channel = dev->btchat_channel;
								
								bt_notify_net(globals, new_node, -1);
								ret_val = 2;
							}
							break;
						case 0:	/* new node */
							#ifdef DEBUG_INFO
							ba2str(&node.addr, addr_str);
							printf("bt_discover_client(): unknown node dev=%d bdaddr=%s\n", i, addr_str);
							#endif
							
							result_node->data = (struct btchat_node_t*) malloc(sizeof(struct btchat_node_t));
							new_node = (struct btchat_node_t*) result_node->data;

							memset(new_node, 0, sizeof(struct btchat_node_t));
							bacpy(&new_node->addr, &node.addr);
							new_node->discovery_method = dev->discovery_method;
							new_node->btchat_channel = dev->btchat_channel;
							new_node->trust_level = BTCHAT_TRUST_LEVEL_NEUTRAL; /* neutral */
							memcpy(new_node->name, dev->name, 250);
							new_node->device_class = dev->device_class;
							new_node->state = dev->state;
							
							bt_notify_net(globals, new_node, -1);
							ret_val = 2;
							break;
						default: /* error ?! */
							#ifdef DEBUG
							printf("bt_discover_client(): res=%d\n", res);
							#endif
							break;
						}							
					} /* for (i < num_devices) */
					
					/* tidy up read buffer */
					memmove(globals->btdiscover.buffer, (globals->btdiscover.buffer + he->length), (globals->btdiscover.buffer_pos - he->length));
					globals->btdiscover.buffer_pos -= he->length;
					/* check buffer again */
					goto check_buffer_again;
				}
				else {
					/* unknown pkg_type - rewind buffer */
					globals->btdiscover.buffer_pos = 0;
				}
			}
			else {
				/* package not complete */
			}
		}
		else {
		}
	} /* fd->revents & POLLIN */
	
	if (event_get(&globals->btdiscover.events, &event) == 1) {
		if (event->etype == BTCHAT_EVENT_DISCOVER) {
			memset(out_buffer, 0, 1024);
			he = (struct btcdiscover_protocol_header_t*) out_buffer;
			dis = (struct btcdiscover_protocol_discover_t*) (out_buffer + sizeof(struct btcdiscover_protocol_header_t));
			
			he->pkg_type = btcdiscover_protocol_discover;
			he->length = sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_discover_t);
			he->time = time(NULL);
			dis->inquiry_time = globals->btdiscover.inquiry_duration;
			
			#ifdef DEBUG_INFO
			printf("bt_discover_client(): inquiry_duration=%d\n", dis->inquiry_time);
			#endif

			res = nwrite(globals->btdiscover.s, out_buffer, he->length);
			#ifdef DEBUG
			printf("bt_discover_client: wrote=%d\n", res);
			#endif
			if (res != he->length) {
				#ifdef DEBUG
				printf("bt_discover_client(): nwrite failed\n");
				#endif
			}
		}
		else {
			#ifdef DEBUG
			printf("bt_discover_client(): unknown event\n");
			#endif
		}
		
		free(event);
		event = NULL;
	}

	return(ret_val);
}

#ifdef OLD_DISCOVER_SERVER
/**
 *	@brief discover other bluetooth devices
 *
 *	This function runs as a separate process (forked of by main)
 *
 *	@param socket	the communication socket
 *
 *	@returns all ways returns 1
 */
int bt_discover_server(int socket)
{
	struct pollfd selectors[1];
	int selectors_ready;
	int read_bytes;
	int write_bytes;
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_discover_t *dis;
	struct btcdiscover_protocol_device_t *dev;
	inquiry_info *info = NULL;
	int flags = 0;
	int num_rsp = 0;
	unsigned char *out_buffer = NULL;
	int i;
	struct btchat_btdiscover_server_t *dglobals;
	char bastr[20];


	dglobals = (struct btchat_btdiscover_server_t*)	malloc(sizeof(struct btchat_btdiscover_server_t));
	
	#ifdef DEBUG
	assert(dglobals != NULL);
	#endif
	
	dglobals->s = socket;
	
	dglobals->buffer_pos = 0;
	dglobals->buffer_size = 1024;
	
	dglobals->bt_device = hci_get_route(NULL);
	
	for (;;) {
		memset(selectors, 0, sizeof(selectors));
		selectors[0].fd = dglobals->s;
		selectors[0].events = POLLIN | POLLPRI;
		/* wait until a command is received from the client */
		selectors_ready = poll(selectors, 1, -1);
						
		/* read and process command */
		if (selectors_ready == 1 && ((selectors[0].revents & POLLIN) == POLLIN || (selectors[0].revents & POLLPRI) == POLLPRI)) {
			#ifdef DEBUG_INFO
			printf("bt_discover_server(): socket is ready to read!\n");
			#endif
			
			read_bytes = cread(dglobals->s, dglobals->buffer+dglobals->buffer_pos, dglobals->buffer_size-dglobals->buffer_pos);
			
			#ifdef DEBUG_INFO
			printf("bt_discover_server(): %d bytes read\n", read_bytes);
			#endif
			
			if (read_bytes == 0) {
				#ifdef DEBUG
				printf("bt_discover_server(): other end closed communication socket, this normally means exit, termination!\n");
				#endif
				break;
			}			
			
			dglobals->buffer_pos += read_bytes;
			
			/* check for protocol header size */
			if (dglobals->buffer_pos >= sizeof(struct btcdiscover_protocol_header_t)) {
				he = (struct btcdiscover_protocol_header_t*) dglobals->buffer;
				
				/* get protocol package type */
				if (he->pkg_type == btcdiscover_protocol_discover || 
						he->pkg_type == btcdiscover_protocol_btchatdiscover) {
					if ((dglobals->buffer_size - sizeof(struct btcdiscover_protocol_header_t))
							>= sizeof(struct btcdiscover_protocol_discover_t)) {
						dis = (struct btcdiscover_protocol_discover_t*) (dglobals->buffer + sizeof(struct btcdiscover_protocol_header_t));
						
						#ifdef DEBUG_INFO
						printf("bt_discover_server(): doing inquiry for %d seconds\n", dis->inquiry_time);
						#endif
						
						/* do bluetooth inquiry */
						num_rsp = hci_inquiry(dglobals->bt_device, dis->inquiry_time, 255, NULL, &info, flags);
						
						/* tidy up buffer - here */
						memmove(dglobals->buffer, (dglobals->buffer+he->length), dglobals->buffer_pos-he->length);
						dglobals->buffer_pos -= he->length;
						
						#ifdef DEBUG_INFO
						printf("bt_discover_server(): buffer_pos=%d\n", dglobals->buffer_pos);
						#endif

						#ifdef DEBUG_INFO
						printf("bt_discover_server(): num devices=%d\n", num_rsp);
						#endif
						
						if (num_rsp <= 0) {
							goto cleanup;
						}
						
						out_buffer = (unsigned char*) malloc(sizeof(struct btcdiscover_protocol_header_t) + (sizeof(struct btcdiscover_protocol_device_t) * num_rsp));
						
						#ifdef DEBUG
						assert(out_buffer != NULL);
						#endif
						
						memset(out_buffer, 0, sizeof(struct btcdiscover_protocol_header_t) + (sizeof(struct btcdiscover_protocol_device_t) * num_rsp));						
						he = (struct btcdiscover_protocol_header_t*) out_buffer;
						he->pkg_type = btcdiscover_protocol_device;
						he->time = time(NULL);
						he->length = sizeof(struct btcdiscover_protocol_header_t) + (sizeof(struct btcdiscover_protocol_device_t) * num_rsp);
						
						if (num_rsp > 0) {
							for (i = 0; i < num_rsp; i++) {
								dev = (struct btcdiscover_protocol_device_t*) (out_buffer + (sizeof(struct btcdiscover_protocol_header_t) + (sizeof(struct btcdiscover_protocol_device_t) * i)));
								memcpy(dev->addr, &(info+i)->bdaddr.b, 6);
								
								/* fixed for testing only */
								ba2str(&(info+i)->bdaddr, dev->name);
								
								dev->device_class = (0x00<<24 & (info+i)->dev_class[0]<<16 & (info+i)->dev_class[1]<<16 & (info+i)->dev_class[2]);
								dev->discovery_method = BTCHAT_DISCOVERY_INQUIRY;
								
								/* fixed channel for testing ONLY */
								dev->btchat_channel = 1;
								
								#ifdef DEBUG_INFO
								ba2str((bdaddr_t*) &dev->addr, bastr);
								printf("bt_discover_server(): dev=%d addr=%s\n", i, bastr);
								#endif
							}
						}
												
						write_bytes = nwrite(dglobals->s, out_buffer, sizeof(struct btcdiscover_protocol_header_t) + (sizeof(struct btcdiscover_protocol_device_t) * num_rsp));
						if (write_bytes == -1) {
							#ifdef DEBUG
							printf("bt_discover_server(): nwrite error\n");
							#endif
						}
						else {
							#ifdef DEBUG
							printf("bt_discover_server(): nwrite=%d\n", write_bytes);
							#endif
						}
					cleanup:
						/* clean up memory */
						if (out_buffer != NULL) {
							free(out_buffer);
							out_buffer = NULL;
						}
						if (info != NULL) {
							free(info);
							info = NULL;
						}
					}
					else {
						continue;
					}
				}
				else if (he->pkg_type == btcdiscover_protocol_null) {
					#ifdef DEBUG_INFO
					printf("bt_discover_server(): protocol_null\n");
					#endif
					continue;
				}
				else {
					#ifdef DEBUG_INFO
					printf("bt_discover_server(): really protocol unknown\n");
					#endif
					continue;
				}
			}
			else {
				#ifdef DEBUG_INFO
				printf("bt_discover_server(): data to small - waiting for more\n");
				#endif
				continue;
			}
		}
		else {
			/* error */
			#ifdef DEBUG
			printf("bt_discover_server(): poll=%d\n", selectors_ready);
			#endif
			continue;
		}
	} /* for (;;) */
	
	return(1);
}
#endif

/**
 *	@brief initialize the new connected client
 *
 *	@param globals	pointer to app. globals
 *	@param index	insertion index for direct connection
 *	@param socket	the connection socket
 *	@param sa	the socket address
 *	@param sa_len	the length of the socket address
 *
 *	@returns 1 on success 2 if client isn't allowed to connect and -1 on error
 */
int bt_client_direct_init(struct btchatd_t *globals, int index, int socket, struct sockaddr_rc *sa, int sa_len)
{
#ifdef OLD_DIRECT_INIT
	struct btchat_node_t current_node;
	struct btchat_node_t *result_node = NULL;
	int res;
	
	
	/* if tree is empty insert new node into tree */
	if (globals->accounts.accounts == NULL) {
		#ifdef DEBUG_INFO
		printf("bt_client_direct_init(): first node\n");
		#endif
	
		globals->accounts.accounts = (struct btchat_accountlist_treenode_t*) malloc(sizeof(struct btchat_accountlist_treenode_t));
		memset(&globals->accounts.accounts, 0, sizeof(struct btchat_accountlist_treenode_t));
		bacpy(&globals->accounts.accounts->account.node.addr, &sa->rc_bdaddr);
		globals->accounts.accounts->account.node.discovery_method = BTCHAT_DISCOVERY_CONNECT;
		globals->accounts.accounts->account.node.btchat_channel = -1;
		globals->accounts.accounts->account.node.trust_level = BTCHAT_TRUST_LEVEL_NEUTRAL; /* neutral */
		globals->accounts.num = 1;
	}

	/* setup search request */
	memset(&current_node, 0, sizeof(struct btchat_accountlist_treenode_t));
	bacpy(&current_node.account.node.addr, &sa->rc_bdaddr);
	
	res = tree_access(globals->accounts.accounts, &current_node, &result_node);
	/* node found */
	if (res == 1) {
		#ifdef DEBUG_INFO
		printf("bt_client_direct_init(): known client\n");
		#endif
	
		/* we don't want a connection to this node! */
		if (result_node->account.node.trust_level <= BTCHAT_TRUST_LEVEL_DONTCONNECT) {
			#ifdef DEBUG_INFO
			printf("bt_client_direct_init(): known client - we don't like connections from it\n");
			#endif
			return(2);
		}
			
		globals->btserver.clients[index].node = result_node;
		globals->btserver.clients[index].s = socket;
		globals->btserver.clients[index].sa_len = sa_len;
		memcpy(&globals->btserver.clients[index].sa, sa, sa_len);
		globals->btserver.clients[index].selector_index = index;
		globals->btserver.clients[index].connection_direction = BTCHAT_CONNECTION_INCOMING;
		globals->btserver.clients[index].state = 1;
		globals->btserver.clients[index].buffer_pos = 0;
		globals->btserver.clients[index].buffer_size = 0xffff;
	}
	/* node not found - new node */
	else if (res == 0) {
		#ifdef DEBUG_INFO
		printf("bt_client_direct_init(): unknown client\n");
		#endif
		
		res = memcmp(&result_node->account.node.addr, &sa->rc_bdaddr, sizeof(bdaddr_t));
		if (res < 0) {
			result_node->left = (struct btchat_accountlist_treenode_t*) malloc(sizeof(struct btchat_accountlist_treenode_t));
			result_node = result_node->left;
		}
		else if (res > 0) {
			result_node->right = (struct btchat_accountlist_treenode_t*) malloc(sizeof(struct btchat_accountlist_treenode_t));
			result_node = result_node->right;
		}
		else {
			#ifdef DEBUG
			printf("bt_client_direct_init(): res==0 in res==0\n");
			#endif
			return(-1);
		}
		/* init new node in tree */
		memset(result_node, 0, sizeof(struct btchat_accountlist_treenode_t));
		bacpy(&result_node->account.node.addr, &sa->rc_bdaddr);
		result_node->account.node.discovery_method = BTCHAT_DISCOVERY_CONNECT;
		result_node->account.node.btchat_channel = -1;
		result_node->account.node.trust_level = BTCHAT_TRUST_LEVEL_NEUTRAL; /* neutral */
		globals->accounts.num++;
		/* init client */
		globals->btserver.clients[index].node = result_node;
		globals->btserver.clients[index].s = socket;
		globals->btserver.clients[index].sa_len = sa_len;
		memcpy(&globals->btserver.clients[index].sa, sa, sa_len);
		globals->btserver.clients[index].selector_index = index;
		globals->btserver.clients[index].connection_direction = BTCHAT_CONNECTION_INCOMING;
		globals->btserver.clients[index].state = 1;
		globals->btserver.clients[index].buffer_pos = 0;
		globals->btserver.clients[index].buffer_size = 0xffff;
	}
	/* error */
	else {
		#ifdef DEBUG
		printf("bt_client_direct_init(): tree_access error\n");
		#endif
		return(-1);
	}
	#endif
	return(1);
}

int bt_client_direct_close(struct btchatd_t *globals, int index)
{
	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	
	close(globals->btserver.clients[index].s);
	globals->btserver.clients[index].s = -1;
	
	return(-1);
}

int bt_notify_net(struct btchatd_t *globals, struct btchat_node_t *node, int index)
{
	struct btchat_event_t *e;
	

	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	
	
	if (index == -1 && node == NULL) return(-1);
	
	e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
	
	#ifdef DEBUG
	assert(e != NULL);
	#endif
	
	memset(e, 0, sizeof(struct btchat_event_t));
	
	if (node == NULL && index != -1) {
		node = globals->btserver.clients[index].node;
	}

	e->etype = BTCHAT_EVENT_NODE;
	memcpy(&e->data.node, node, sizeof(struct btchat_node_t));
	e->prev = NULL;
	
	if (index != -1) {
		e->data.node.connection = globals->btserver.clients[index].connection_direction;
	}
	else {
		e->data.node.connection = 0;
	}
	
	event_put(&globals->net.events, e);
	
	return(1);
}
