/**
 *	@file bt_discover.c
 *
 *	@project btChat
 *
 *	@author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 * (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "node_protocol.h"
#include "event.h"
#include "util.h"
#include "discover_protocol.h"
#include "bt_discover.h"


/**
 *	@brief compare two discover_nodes for searching the tree
 *
 *	@param d1	pointer to left node
 *	@param d2	pointer to right node
 *
 *	@returns 0 if equal, 1 if d1 > d2 and -1 if d2 > d1
 */
static int btd_tree_cmp(const void *d1, const void *d2)
{
	struct btchat_discover_node_t *a1 = (struct btchat_discover_node_t*) d1;
	struct btchat_discover_node_t *a2 = (struct btchat_discover_node_t*) d2;
	
	#ifdef DEBUG
	assert(a1 != NULL);
	assert(a2 != NULL);
	#endif
	
	return(memcmp(a1->addr.b, a2->addr.b, 6));
}

/**
 *	@brief send devinfo to discover client if given node has changed status
 *
 *	@param node	pointer to the node to check
 *	@param data	pointer to bt_discover_server2 globals
 *
 *	@returns 1
 */
static int btd_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_discover_node_t *n = NULL;
	struct btchat_btdiscover_server_t *g = NULL;
	unsigned char out_buffer[1024];
	int out_buffer_size = 1024;
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_device_t *dev;
	int write_bytes = 0;
	
	
	#ifdef DEBUG
	assert(node != NULL);
	assert(data != NULL);
	#endif


	n = (struct btchat_discover_node_t*) node->data;
	g = (struct btchat_btdiscover_server_t*) data;

	/* detect state change! */
	if (n->last < g->current_time && n->state != BTCHAT_NODE_STATE_OFFLINE) {
		n->state = BTCHAT_NODE_STATE_OFFLINE;
	}
	else if (n->updated == 1) {
	}
	else if (n->state == BTCHAT_NODE_STATE_OFFLINE && n->last == g->current_time) {
		n->state = BTCHAT_NODE_STATE_ONLINE;
	}
	else {
		#ifdef DEBUG_INFO
		printf("btd_tree_walk(): nothing happened name=%s\n", n->name);
		#endif
		/* nothing happened */
		return(1);
	}

	memset(out_buffer, 0, out_buffer_size);
	he = (struct btcdiscover_protocol_header_t*) out_buffer;
	he->pkg_type = btcdiscover_protocol_device;
	he->time = g->current_time;
	he->length = (sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_device_t));
						
	dev = (struct btcdiscover_protocol_device_t*) (out_buffer + sizeof(struct btcdiscover_protocol_header_t));
	memcpy(dev->addr, n->addr.b, 6);
								
	memcpy(dev->name, n->name, 250);
		
	dev->device_class = n->device_class;
	dev->discovery_method = n->discovery_method;
								
	dev->btchat_channel = n->btchat_channel;
			
	/* reset updated flag */
	n->updated = 0;
	dev->state = n->state;
		
	#ifdef DEBUG_INFO
	printf("btd_tree_walk(): name=%s\n", dev->name);
	#endif
		
	write_bytes = nwrite(g->s, out_buffer, sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_device_t));
	if (write_bytes == -1) {
	#ifdef DEBUG
		printf("btd_tree_walk(): nwrite error\n");
		#endif
			
		/* break tree walk on send error */
		return(2);
	}
	else {
		#ifdef DEBUG
		printf("btd_tree_walk(): nwrite=%d\n", write_bytes);
		#endif
	}

	return(1);
}

/**
 *	@brief discover other bluetooth devices
 *
 *	This function runs as a separate process (forked of by main)
 *
 *	@param socket	the communication socket
 *
 *	@returns allays returns 1
 */
int bt_discover_server2(int socket)
{
	struct btchat_btdiscover_server_t *dg = NULL;
	struct pollfd selectors[1];
	int selectors_ready;
	int read_bytes;
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_discover_t *dis;
	inquiry_info *info = NULL;
	int flags = 0;
	int num_rsp = 0;
	int i;
	#ifdef DEBUG
	char bastr[20];
	#endif
	char tmp_name[250];
	int res;
	struct btchat_discover_node_t node;
	struct btchat_discover_node_t *new_node;
	struct tree_node_t *result_node;
	int hci_dev;
	

	/* allocate globals */
	dg = (struct btchat_btdiscover_server_t*)	malloc(sizeof(struct btchat_btdiscover_server_t));
	
	#ifdef DEBUG
	assert(dg != NULL);
	#endif
	
	dg->s = socket;
	
	dg->buffer_pos = 0;
	dg->buffer_size = 1024;
	
	dg->bt_device = hci_get_route(NULL);
	
	/* init b-tree */
	dg->btdevices.root = NULL;
	dg->btdevices.num = 0;
	dg->btdevices.compare_function = btd_tree_cmp;
	dg->btdevices.walk_function = btd_tree_walk;

	/* bt_discover_server2() main loop */	
	for (;;) {
		memset(selectors, 0, sizeof(selectors));
		selectors[0].fd = dg->s;
		selectors[0].events = POLLIN | POLLPRI;
		
		/* wait until a command is received from the client */
		selectors_ready = poll(selectors, 1, -1);
		
		/* check poll(..) results */
		if (selectors_ready != 1) {
			continue;
		}
		if ((selectors[0].revents & POLLIN) != POLLIN && (selectors[0].revents & POLLPRI) != POLLPRI) {
			continue;
		}
		
		#ifdef DEBUG_INFO
		printf("bt_discover_server2(): socket is ready to read!\n");
		#endif
			
		read_bytes = cread(dg->s, (dg->buffer + dg->buffer_pos), (dg->buffer_size - dg->buffer_pos));
			
		#ifdef DEBUG_INFO
		printf("bt_discover_server2(): %d bytes read\n", read_bytes);
		#endif
			
		if (read_bytes == 0) {
			#ifdef DEBUG
			printf("bt_discover_server2(): other end closed communication socket, this normally means exit, termination!\n");
			#endif
			break;
		}			
		
		/* advance buffer position */
		dg->buffer_pos += read_bytes;
			
		/* check for protocol header size */
		if (dg->buffer_pos >= sizeof(struct btcdiscover_protocol_header_t)) {
		
			he = (struct btcdiscover_protocol_header_t*) dg->buffer;
				
			/* pkg_type == discover devices */
			if (he->pkg_type == btcdiscover_protocol_discover || he->pkg_type == btcdiscover_protocol_btchatdiscover) {
				if ((dg->buffer_size - sizeof(struct btcdiscover_protocol_header_t)) >= sizeof(struct btcdiscover_protocol_discover_t)) {
					
					dis = (struct btcdiscover_protocol_discover_t*) (dg->buffer + sizeof(struct btcdiscover_protocol_header_t));
						
					#ifdef DEBUG_INFO
					printf("bt_discover_server2(): doing inquiry for %d seconds\n", dis->inquiry_time);
					#endif
						
					/* do bluetooth inquiry */
				inquiry:
					num_rsp = hci_inquiry(dg->bt_device, dis->inquiry_time, 255, NULL, &info, flags);
					
					if (num_rsp < 0) {
						#ifdef DEBUG
						printf("bt_discover_server2(): hci_inquiry() failed ... try to reinitialize hci device\n");
						sleep(10);
						#endif
						
						dg->bt_device = hci_get_route(NULL);
						goto inquiry;
					}
					
					/* tidy up read buffer - here */
					memmove(dg->buffer, (dg->buffer + he->length), (dg->buffer_pos - he->length));
					dg->buffer_pos -= he->length;
						
					#ifdef DEBUG_INFO
					printf("bt_discover_server2(): buffer_pos=%d\n", dg->buffer_pos);
					#endif

					#ifdef DEBUG_INFO
					printf("bt_discover_server2(): num devices=%d\n", num_rsp);
					#endif
					
					/* get current time */
					time(&dg->current_time);
					
					/* open hci device for name requests */
					hci_dev = hci_open_dev(dg->bt_device);
					if (hci_dev < 0) {
						perror("bt_discover_server2() hci_open_dev()");
					}
					
					/* process inquiry results */
					for (i = 0; i < num_rsp; i++) {
						memset(&node, 0, sizeof(struct btchat_discover_node_t));
						memcpy(node.addr.b, &(info+i)->bdaddr.b, 6);
							
						#ifdef DEBUG_INFO
						ba2str(&node.addr, bastr);
						printf("bt_discover_server2(): dev=%d addr=%s\n", i, bastr);
						#endif
							
						result_node = NULL;
						res = tree_access_insert(&dg->btdevices, (void*) &node, &result_node);
						
						switch (res) {
						case 1:	/* known node */
							/* mark last contact */
							new_node = (struct btchat_discover_node_t*) result_node->data;
							
							#ifdef DEBUG_INFO
							printf("bt_discover_server2() known node state=%d\n",new_node->state);
							#endif
							if (new_node->state == BTCHAT_NODE_STATE_OFFLINE) {
								memset(tmp_name, 0, 250);
								res = hci_read_remote_name(hci_dev, &(info+i)->bdaddr, 250, tmp_name, 2);
								if (res < 0) {
									#ifdef DEBUG
									perror("bt_discover_server2() hci_read_remote_name()");
									#endif
								}
								else {
									memcpy(new_node->name, tmp_name, 250);
									/*new_node->updated = 1;*/
								}
								#ifdef DEBUG_INFO
								printf("bt_discover_server2() known name=%s\n", new_node->name);
								#endif
							}
							new_node->last = dg->current_time;
							break;
						case 0:	/* new node */
							#ifdef DEBUG_INFO
							ba2str(&node.addr, bastr);
							printf("bt_discover_server2(): unknown node dev=%d bdaddr=%s\n", i, bastr);
							#endif
					
							result_node->data = malloc(sizeof(struct btchat_discover_node_t));
							new_node = (struct btchat_discover_node_t*) result_node->data;
							memset(new_node, 0, sizeof(struct btchat_discover_node_t));
							
							bacpy(&new_node->addr, &node.addr);
							new_node->discovery_method = BTCHAT_DISCOVERY_INQUIRY;
							new_node->trust_level = 0;
							
							/* fixed for testing only */
							new_node->btchat_channel = -1;
							
							/* get bt device name */
							res = hci_read_remote_name(hci_dev, &(info+i)->bdaddr, 250, new_node->name, 2);
							if (res < 0) {
								#ifdef DEBUG
								perror("bt_discover_server2() hci_read_remote_name()");
								#endif
								ba2str(&(info+i)->bdaddr, new_node->name);
							}
							#ifdef DEBUG_INFO
							printf("bt_discover_server2() name=%s\n",new_node->name);
							#endif
							
							new_node->last = dg->current_time;
							new_node->updated = 1;
							new_node->state = BTCHAT_NODE_STATE_ONLINE;
							
							new_node->device_class = (0x00<<24 & (info+i)->dev_class[0]<<16 & (info+i)->dev_class[1]<<16 & (info+i)->dev_class[2]);
							break;
						default:
							#ifdef DEBUG_INFO
							printf("bt_discover_server2(): res unknown - ERROR\n");
							#endif
							break;
						}
					} /* for (i < num_rsp) */
					
					/* close hci device */
					if (hci_dev > 0) {
						close(hci_dev);
					}
					
					/* free inquiry information */
					if (info != NULL) {
						free(info);
						info = NULL;
					}

					/* check all nodes for change */
					if (tree_walk(&dg->btdevices, dg) == 2) {
						#ifdef DEBUG
						printf("bt_discover_server2(): tree_walk reports: lost connection, terminating!\n");
						break;
						#endif
					}
				}
				else {
					#ifdef DEBUG_INFO
					printf("bt_discover_server2(): body data to small - waiting for more\n");
					#endif
					continue;
				}
			}
			else if (he->pkg_type == btcdiscover_protocol_null) {
				#ifdef DEBUG_INFO
				printf("bt_discover_server2(): protocol_null\n");
				#endif
				continue;
			}
			else {
				#ifdef DEBUG_INFO
				printf("bt_discover_server2(): really protocol unknown\n");
				#endif
				continue;
			}
		}
		else {
			#ifdef DEBUG_INFO
			printf("bt_discover_server2(): header data to small - waiting for more\n");
			#endif
			continue;
		}
	} /* for (;;) */
	
	return(1);
}
