/**
 *	@file btchatd.h
 *
 *	@project btChat
 *
 *	@author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 * (c) Collin R. Mulliner
 *
 */

#ifndef __BTCHATD_H__
#define __BTCHATD_H__

enum trust_level_t {
	BTCHAT_TRUST_LEVEL_DONTCONNECT = -2,
	BTCHAT_TRUST_LEVEL_NEUTRAL = 0,
	BTCHAT_TRUST_LEVEL_ROUTE
};

enum discovery_method_t {
	BTCHAT_DISCOVERY_NULL = 0,
	BTCHAT_DISCOVERY_INQUIRY,
	BTCHAT_DISCOVERY_CONNECT,
	BTCHAT_DISCOVERY_NET,
	BTCHAT_DISCOVERY_LAST = 255
};

enum connection_direction_t {
	BTCHAT_CONNECTION_NONE = 0,
	BTCHAT_CONNECTION_INCOMING = 1,
	BTCHAT_CONNECTION_OUTGOING = 2
};

enum btchat_node_state_t {
	BTCHAT_NODE_STATE_FIRST = 0,
	BTCHAT_NODE_STATE_ONLINE,
	BTCHAT_NODE_STATE_OFFLINE,
	BTCHAT_NODE_STATE_AWAY,
	BTCHAT_NODE_STATE_VISIBLE,
	BTCHAT_NODE_STATE_LAST = 255
};

enum btchat_route_policy_t {
	BTCHAT_ROUTE_POLICY_FIRST = 0,
	BTCHAT_ROUTE_POLICY_NEVER,	/**< do not route any traffic */
	BTCHAT_ROUTE_POLICY_TRUSTED,	/**< only route traffic from trusted devices - default */
	BTCHAT_ROUTE_POLICY_NEEDUSER,	/**< only route if user is active */
	BTCHAT_ROUTE_POLICY_LAST = 255
};

struct btchat_node_t {

	bdaddr_t addr;	/**< the nodes bluetooth device address */
	unsigned int device_class;	/**< the nodes bluetooth device service class */
	char name[250];	/**< the nodes bluetooth device name */
	
	short int btchat_channel;	/**< the btchat rfcomm channel */

	short int discovery_method;	/**< method of device discovery */
	
	int trust_level;	/**< level of trust */
	
	int state;	/**< state */
	
	char alias[250];	/**< the visible name */
	
	time_t last;	/**< time of last contact with node */
};

enum btchat_event_type_t {
	BTCHAT_EVENT_NULL = 0,
	BTCHAT_EVENT_DISCOVER,
	BTCHAT_EVENT_NODE,
	BTCHAT_EVENT_MSG,
	BTCHAT_EVENT_LASTEVENT = 255
};

/**
 *	@brief a btchat event
 */
struct btchat_event_t {

	int etype;	/**< event type */
	
	union {
		struct {
			struct btchat_node_t node;
			int connection;
		} node;
	} data;
	
	struct btchat_event_t *prev;	/**< previous event in queue */
};

/**
 *	@brief event queue
 *
 */
struct btchat_event_queue_t {

	struct btchat_event_t *head;	/**< first (newest) event in queue */
	struct btchat_event_t *tail;	/**< last (oldest) event in queue */
};

struct btchat_btdiscover_client_t {

	int s;	/**< socket */
	
	int pid;	/**< the child pid */
	
	int inquiry_duration;	/**< bt inquiry duration */
	int local_monitor;	/**< yes=1 no=0 */
	int local_monitor_interval;	/**< inquiry every X seconds */
	time_t local_monitor_last;	/**< last inquiry time */
	
	unsigned char buffer[0xffff];	/**< receive buffer */
	int buffer_pos;	/**< position in buffer */
	int buffer_size;	/**< size of buffer */
	
	struct btchat_event_queue_t events;	/**< events */
};

struct btchat_btdirect_t {

	struct btchat_node_t *node;	/**< the actual node */
	
	int connection_direction;	/**< role of connection, incoming or outgoing */
	
	int s;	/**< the connection socket */
	struct sockaddr_rc sa;	/**< the socket info */
	int sa_len;
	
	unsigned char buffer[0xffff];
	int buffer_pos;
	int buffer_size;
	
	int state;	/**< current state */
	
	int selector_index;	/**< index for global event_loop */
	
	struct btchat_event_queue_t events;	/**< events */
};

struct btchat_btserver_t {

	int server;	/**< server socket */
	struct sockaddr_rc server_sa;
	int server_len;

	int route_policy;	/**< */

	struct btchat_btdirect_t clients[7];	/**< clients */
	int clients_num;	/**< number of clients */
	int clients_max;	/**< max number of clients to connect */
	
	int state;	/**< current state */
	
	short int channel;	/**< rfcomm channel to listen on */
	bdaddr_t local_bdaddr;	/**< local bdaddr to bind to */
	
	struct btchat_event_queue_t events;	/**< events which don't have a active connection jet */
};

struct btchat_network_t {

	short int port;	/**< TCP port to listen on */
	char local_ip[256];	/**< ip to bind to */

	struct hostent *hostinfo;
	struct sockaddr_in laddr;
	struct sockaddr_in remote_sa;	/**< client socket information */
	socklen_t remote_len;	/**< size of remote_sa */

	int server;	/**< server socket */
	int client;	/**< client socket */

	unsigned char buffer[0xffff];
	int buffer_pos;
	int buffer_size;

	int state;
	
	struct btchat_event_queue_t events;	/**< events */
};

struct btchat_cron_t {

	int interval;
	time_t next;
	time_t last;
};

struct btchatd_t {
	struct btchat_node_t self;
	
	struct btchat_network_t net;
	struct btchat_btserver_t btserver;
	struct btchat_btdiscover_client_t btdiscover;

	struct tree_t nodes;

	struct btchat_cron_t tasks;

	int running;	/**< global running flag */
};

#endif
