/**
 *	@file net.c
 *
 *	@project btChat
 *
 *	@author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 * (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "event.h"
#include "util.h"
#include "client_protocol.h"
#include "net.h"

/**
 *	@brief handle the connection with the user front-end
 *
 *	This function binds to the given local ip-address and port, normally
 *	127.0.0.1 and accepts exactly one (1) connection.
 *
 *	@param args	pointer to globals
 *	@param fd	pointer to file descriptor
 *
 *	@returns 1 on success and -1 on failure
 */
int net_server(struct btchatd_t *globals, struct pollfd *fd)
{
	int read_bytes;
	int write_bytes;
	struct btchat_event_t *event;
	unsigned char *out_buffer = NULL;
	unsigned int out_buffer_size;
	unsigned int out_buffer_pos;
	struct btcclient_protocol_account_t *dev;
	struct btcclient_protocol_header_t *he;
	

	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	
	
	/*
	 *	state == 0 => network initialization
	 */
	if (globals->net.state == 0) {
	
		globals->net.laddr.sin_port = htons((u_short) globals->net.port);
		if ((globals->net.hostinfo = gethostbyname(globals->net.local_ip))) {
			bcopy(globals->net.hostinfo->h_addr, (char *) &globals->net.laddr.sin_addr, globals->net.hostinfo->h_length);
			globals->net.laddr.sin_family = AF_INET;
		}
		else {
			#ifdef DEBUG
			perror("net_server(): gethostbyname failed");
			#endif
			return(-1);
		}
	
		if ((globals->net.server = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
			#ifdef DEBUG
			perror("net_server(): socket failed");
			#endif
			return(-1);
		}
	
		if (bind(globals->net.server, (struct sockaddr*) &globals->net.laddr, sizeof(struct sockaddr_in)) == -1) {
			#ifdef DEBUG
			perror("net_server(): bind failed");
			#endif
			return(-1);
		}
	
		if (listen(globals->net.server, 1) == -1) {
			#ifdef DEBUG
			perror("net_server(): listen failed");
			#endif
			return(-1);
		}
		
		globals->net.state = 1;
		return(1);
	} /* state == 0 */
	/*
	 *	state == 1 => accepting new client connection
	 */
	else if (globals->net.state == 1) {
		if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
			/* accept incoming client connection */
			if (fd->fd == globals->net.server) {
				globals->net.client = accept(globals->net.server, (struct sockaddr*) &globals->net.remote_sa, &globals->net.remote_len);
				
				/* close server socket */
				close(globals->net.server);
				globals->net.server = -1;
				
				/* we have a client now */
				globals->net.state = 2;
				
				globals->net.buffer_pos = 0;
				
				#ifdef DEBUG_INFO
				printf("net_server(): accepted new connection\n");
				#endif
				
				net_notify_con(globals);
				
				#ifdef DEBUG_INFO
				printf("net_server(): sent current nodes to client\n");
				#endif
			}
			else {
				#ifdef DEBUG_INFO
				printf("net_server(): server socket corrupted, going to state=0\n");
				#endif
				
				close(globals->net.server);
				globals->net.server = -1;
				globals->net.state = 0;
				return(net_server(globals, NULL));
			}
		}
	}
	/*
	 *	state == 2 => talk to client
	 */
	else if (globals->net.state == 2) {
		if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
			/* read data from client */
			if (fd->fd == globals->net.client) {
			
				read_bytes = cread(globals->net.client, globals->net.buffer + globals->net.buffer_pos, globals->net.buffer_size - globals->net.buffer_pos);
			
				/* client disconnected */
				if (read_bytes == 0) {
					#ifdef DEBUG
					printf("net_server(): read_bytes=0, looks like the user client has closed the connection\n");
					#endif
					
					close(globals->net.client);
					globals->net.client = -1;
					globals->net.state = 0;
					event_queue_reset(&globals->net.events);
					
					/* listen for incoming connections */
					return(net_server(globals, NULL));
				}
				
				globals->net.buffer_pos += read_bytes;
				
				/* do protocol action here ! */
			
			}
		} /* POLLIN || POLLPRI */

		/* process pending events */		
		if (event_get(&globals->net.events, &event) == 1) {
		
			/* node information event */
			if (event->etype == BTCHAT_EVENT_NODE) {
				out_buffer_size = sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_account_t);
				out_buffer_pos = 0;
				out_buffer = (unsigned char*) malloc(out_buffer_size);
				
				#ifdef DEBUG
				assert(out_buffer != NULL);
				#endif
				
				memset(out_buffer, 0, out_buffer_size);
				
				he = (struct btcclient_protocol_header_t*) out_buffer;
				dev = (struct btcclient_protocol_account_t*) (out_buffer + sizeof(struct btcclient_protocol_header_t));
				
				he->pkg_type = BTCCLIENT_PROTOCOL_ACCOUNT;
				time((time_t*) &he->time);
				he->length = out_buffer_size;
				
				/* memcpy(dev->addr, event->data.node.node.addr.b, 6); */
				ba2str(&event->data.node.node.addr, dev->addr);
				dev->device_class = event->data.node.node.device_class;
				memcpy(dev->name, event->data.node.node.name, 250);				
				dev->btchat_channel = event->data.node.node.btchat_channel;
				dev->discovery_method = event->data.node.node.discovery_method;
				dev->trust_level = event->data.node.node.trust_level;
				dev->connection = event->data.node.connection;
				dev->state = event->data.node.node.state;
				
				#ifdef DEBUG_INFO
				printf("net_server(): BTCHAT_EVENT_NODE name=%s\n", dev->name);
				#endif
			}
			/* unknown event - free and return */
			else {
				event_free(&event);
				return(1);
			}
			
			write_bytes = nwrite(globals->net.client, out_buffer, out_buffer_size);
			#ifdef DEBUG
			printf("net_server(): written bytes=%d\n", write_bytes);
			#endif
			
			if (write_bytes == 0) {
				event_free(&event);
				free(out_buffer);
				out_buffer = NULL;
				
				#ifdef DEBUG
				printf("net_server(): write_bytes=0, looks like the user client has closed the connection\n");
				#endif
					
				close(globals->net.client);
				globals->net.client = -1;
				globals->net.state = 0;
				event_queue_reset(&globals->net.events);
					
				/* listen for incoming connections */
				return(net_server(globals, NULL));
			}
		
			/* cleanup */
			if (out_buffer != NULL) free(out_buffer);
			out_buffer = NULL;
			event_free(&event);
		}
	} /* state == 2 */
	
	/* clear event queue - no client to consume events */
	if (globals->net.state != 2) {
		#ifdef DEBUG_INFO
		/*printf("net_server(): not connected, clearing event queue\n");*/
		#endif
		event_queue_reset(&globals->net.events);
		return(1);
	}
	
	if (event_available(&globals->net.events) == 1) return(2);
	else return(1);
}

/**
 *	@brief if node is online notify user client (net)
 *
 *	This is a tree_walk callback.
 *
 *	@param node	pointer to the node to check
 *	@param data	pointer to globals
 *
 *	@returns 1
 */
static int net_notify_con_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_node_t *n = NULL;
	struct btchatd_t *g = NULL;
	struct btchat_event_t *e = NULL;
	
	
	#ifdef DEBUG
	assert(node != NULL);
	assert(data != NULL);
	#endif


	n = (struct btchat_node_t*) node->data;
	g = (struct btchatd_t*) data;
	
	if (n->state != BTCHAT_NODE_STATE_ONLINE) {
		return(1);
	}
	
	#ifdef DEBUG_INFO
	printf("net_notify_con_tree_walk(): checking name=%s state=%d\n", n->name, n->state);
	#endif
	
	e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
	
	#ifdef DEBUG
	assert(e != NULL);
	#endif
	
	memset(e, 0, sizeof(struct btchat_event_t));
	
	e->etype = BTCHAT_EVENT_NODE;
	memcpy(&e->data.node, n, sizeof(struct btchat_node_t));
	e->prev = NULL;
	e->data.node.connection = 0;

	event_put(&g->net.events, e);

	return(1);
}

/**
 *	@brief notify fresh connected user client about online nodes
 *
 * @pram globals pointer to globals
 *
 * @returns all ways 1
 */
int net_notify_con(struct btchatd_t *globals)
{
	void *old_walk;


	#ifdef DEBUG
	assert(globals != NULL);
	#endif
	
	/* save old walk function */
	old_walk = globals->nodes.walk_function;
	/* install new walk function */
	globals->nodes.walk_function = net_notify_con_tree_walk;
	
	/* walk tree */
	tree_walk(&globals->nodes, globals);
	
	/* restore old walk function */
	globals->nodes.walk_function = old_walk;
	
	return(1);
}
