/**
 *	@file tree.c
 *
 *	@project btChat
 *
 *	@author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 * (c) Collin R. Mulliner
 *
 */

#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include "tree.h"

static int tree_walk_search(struct tree_node_t *cur, tree_walk_callback *walk, void *user)
{
	/* go left */
	if (cur->l != NULL)
		/* 2 == interrupt - break walk */
		if (tree_walk_search(cur->l, walk, user) == 2) return(2);
		
	/* go right */
	if (cur->r != NULL)
		/* 2 == interrupt - break walk */
		if (tree_walk_search(cur->r, walk, user) == 2) return(2);
	
	/* do user action for current node */
	return(walk(cur, user));
}

int tree_walk(const struct tree_t *tree, void *user)
{
	if (tree->root == NULL) return(1);

	return(tree_walk_search(tree->root, tree->walk_function, user));
}

static int tree_access_search(struct tree_node_t *cur, const void *search, struct tree_node_t **result, tree_compare_callback *cmp)
{
	int res;

	
	#ifdef DEBUG
	assert(search != NULL);
	assert(result != NULL);
	assert(cmp != NULL);
	#endif
	

	if (cur != NULL) {
		res = cmp(cur->data, search);
		
		if (res < 0) {
			if (cur->l == NULL) {
				*result = cur;
				return(0);
			}
			else {
				return(tree_access_search(cur->l, search, result, cmp));
			}
		}
		else if (res > 0) {
			if (cur->r == NULL) {
				*result = cur;
				return(1);
			}
			else {
				return(tree_access_search(cur->r, search, result, cmp));
			}
		}
		else {
			*result = cur;
			return(2);
		}
	}
	else {
		return(-1);
	}
}

/**
 *	@brief find search element in tree
 *
 *	the result is written to result
 *
 *	@param tree   pointer to tree
 *	@param search pointer to search data (interpreted by compare function)
 *	@param result pointer to pointer for result
 *
 *	@returns -1 on parameter error, 0 if search failed and 1 on success
 */
int tree_access(const struct tree_t *tree, const void *search, struct tree_node_t **result)
{
	int res;
	

	if (tree == NULL) return(-1);
	if (search == NULL) return(-1);
	if (result == NULL) return(-1);

	if (tree->root == NULL) return(0);

	res = tree_access_search(tree->root, search, result, tree->compare_function);
	
	switch (res) {
	case 2:
		return(1);
	case -1:
		return(-1);
	default:
		return(0);
	}
}

/**
 *	@brief find search element in tree or create new element
 *
 *	The result is written to result, when the search fails (with result == 0)
 * the new allocated node is return as result.
 *
 *	@param tree   pointer to tree
 *	@param search pointer to search data (interpreted by compare function)
 *	@param result pointer to pointer for result
 *
 *	@returns -1 on parameter error, 0 if new node was allocated and 1 on success
 */
int tree_access_insert(struct tree_t *tree, const void *search, struct tree_node_t **result)
{
	int res;
	

	if (tree == NULL) return(-1);
	if (search == NULL) return(-1);
	if (result == NULL) return(-1);

	if (tree->root == NULL) {
		tree->root = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		memset(tree->root, 0, sizeof(struct tree_node_t));
		tree->num = 1;
		*result = tree->root;
		return(0);
	}

	res = tree_access_search(tree->root, search, result, tree->compare_function);
	
	if (res == 0) {
		(*result)->l = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		*result = (*result)->l;
	}
	else if (res == 1) {
		(*result)->r = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		*result = (*result)->r;
	}
	else {
		return((res == 2 ? 1 : res));
	}
	
	memset((*result), 0, sizeof(struct tree_node_t));
	tree->num++;
	return(0);
}
